{-
/***************************************************************************
 *   Copyright (C) 2004 by Peter Robinson,listener@thaldyron.com           *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         * 
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
-}

module InputParser 
where

import Data.List
import Definitions

data ParseRes = AddFile String
              | RemFile String
              | UpdFile String
              | Skip
              | ParseError String
              
parseInput :: String -> ParseRes
parseInput (c:fname) 
        | (isIndexableFile fname) && (c == 'a')  = AddFile ('/':fname)
        | (isIndexableFile fname) && (c == 'r')  = RemFile ('/':fname)
        | (isIndexableFile fname) && (c == 'u')  = UpdFile ('/':fname)        
        | otherwise                              = Skip

        where
        stripPath :: String -> String
        stripPath path
                | '/' `elem` path  = stripPath ((tail . dropWhile (/='/')) path)
                | otherwise        = path
                
        isIndexableFile :: String -> Bool
        isIndexableFile fname = (not (
                                   (".glimpse" `isPrefixOf` (stripPath fname)) || -- no files of glimpse
                                   ("/." `isSublist` (stripPath fname))        || -- no hidden files 
                                   ("~" `isSublist` (stripPath fname))            -- no backup files
                                     )
                                ) 
                                && 
                                ((tail indexRootDir) `isPrefixOf` fname)   -- only files below index root
                                         
        
parseInput err         = ParseError err               


isSublist p = or . map (isPrefixOf p) . tails