/***************************************************************************
 *   Implementation as a LSM for Kernel 2.6                                *
     Blocking- and nonblocking device implementation, can be used as a FIFO*
     Copyright (C) 2004 by Peter Robinson,listener@thaldyron.com           *
 
 *   Parts of the device- and struct filenames_list handling function      *
     were taken from rlocate (http://stud3.tuwien.ac.at/~e9526925/)        *
     Copyright (C) 2004 by Rastislav Levrinc                               *    
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         * 
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/string.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/fs.h>
#include <linux/mount.h>
#include <linux/namei.h>
#include <linux/namespace.h>
#include <linux/limits.h>
#include <linux/security.h>
#include <linux/list.h>

#define __NO_VERSION__
#define MY_NAME "fsysmon"
#define DEVICE_NAME "fsysmon"
#define DEVICE_MINOR 1
#define MODULE_NAME "fsysmon"
#define AUTHOR "Peter Robinson"
#define LICENSE "GPL"


struct filenames {
        struct list_head list;
        char *msg;
};

LIST_HEAD(filenames_list);


MODULE_DESCRIPTION( MODULE_NAME );
MODULE_AUTHOR( AUTHOR );

MODULE_LICENSE( LICENSE );


/*  PROTOTYPES         */
//static char *getFullPath( struct dentry *dentry );
// char *build_path_from_dentry(struct dentry *direntry);
void error_occurred( const char *msg );
void register_device( void );
void register_as_security_module( void );
void unregister_device( void );
void unregister_as_security_module( void );
inline static void remove_all_filenames_entries( void );
inline static void remove_filenames_entry(struct filenames *filenames_entry);
inline static void filenames_entry(const char *path, char mode );
// static void filenames_entry_a(const char *filename);


// void override_syscalls( void );
// void restore_syscalls( void );

/* device ops                       */
loff_t fsysmon_llseek( struct file *, loff_t, int );
ssize_t fsysmon_read( struct file *, char *, size_t, loff_t * );
ssize_t fsysmon_write( struct file *, const char *, size_t, loff_t * );
int fsysmon_ioctl( struct inode *, struct file *, unsigned int, unsigned long );
int fsysmon_open( struct inode *, struct file * );
int fsysmon_release( struct inode *, struct file * );
unsigned int fsysmon_poll( struct file *, poll_table *);

//char * d_path(struct dentry *dentry, struct vfsmount *vfsmnt,
//                                char *buf, int buflen);
inline static struct vfsmount *find_vfsmount(struct dentry *dentry) ;
/*static char * __d_path( struct dentry *dentry, struct vfsmount *vfsmnt,
                        struct dentry *root, struct vfsmount *rootmnt,
                        char *buffer, int buflen);*/
static int fsysmon_inode_create(struct inode *dir, struct dentry *dentry, int mode);
static int fsysmon_inode_mkdir( struct inode * dir,struct dentry * dentry, int mode );
static int fsysmon_inode_rmdir( struct inode * dir, struct dentry * dentry );
static int fsysmon_inode_unlink( struct inode * dir, struct dentry * dentry );
static int fsysmon_inode_rename( struct inode * old_dir,struct dentry * old_dentry,
                                 struct inode * new_dir,struct dentry * new_dentry );
static int fsysmon_sb_mount( char *dev_name, struct nameidata * nd,char *type, unsigned long flags, void *data );
static int fsysmon_sb_umount( struct vfsmount * mnt, int flags );
static int fsysmon_inode_permission( struct inode * inode, int mask, struct nameidata *nd );
// int (*orig_mkdir) ( const char *path, mode_t mode ) = NULL;
// int (*orig_creat) ( const char *path, mode_t mode ) = NULL;
// int (*orig_rename) ( const char *oldpath, const char *newpath ) = NULL;
// int (*orig_rmdir) ( const char *pathname ) = NULL;
// int (*orig_unlink) ( const char *pathname ) = NULL;
// int (*orig_umount) ( const char *target ) = NULL;
// int (*orig_mknod) ( const char *path, mode_t mode, dev_t dev ) = NULL;
// int (*orig_mount) ( const char *source, const char *target, 
//                    const char *filesystemtype, unsigned long mountflags, 
//                    const void *data ) = NULL;




/* *************************          DATA  *********************       */

/* flag to keep track of how we were registered as a security module*/
static int secondary;

// static struct vfsmount *globalmnt;

static int major = 0;
static char *tmp_path;      
static char *tmp_path2;      

static struct file_operations fsysmon_ops = {
    llseek: fsysmon_llseek,
    read: fsysmon_read,
    write: fsysmon_write,
    ioctl: fsysmon_ioctl,
    open: fsysmon_open,
    release: fsysmon_release,
    poll: fsysmon_poll,
    owner: THIS_MODULE,
};


static struct security_operations fsysmon_security_ops = {
        .inode_create =                 fsysmon_inode_create,
        .inode_mkdir =                  fsysmon_inode_mkdir,
        .inode_rmdir =                  fsysmon_inode_rmdir,
        .inode_unlink =                 fsysmon_inode_unlink,
        .inode_rename =                 fsysmon_inode_rename,
        .inode_permission =             fsysmon_inode_permission,
        .sb_mount =                     fsysmon_sb_mount,
        .sb_umount =                    fsysmon_sb_umount, 
};

static wait_queue_head_t inq;




/* **************************** SECURITY MODULE HOOKS ******************************* */


// finds the corresponding vfsmounts for a dentry automagically by traversing the entire namespace
// Not very efficient but at the moment i don't have a better solution for kernel 2.6
inline static struct vfsmount *find_vfsmount(struct dentry *dentry) 
{ 
         struct dentry *sb_root = dget(dentry->d_sb->s_root); 
         struct namespace *namespace = current->namespace; 
         struct list_head *head; 
         struct vfsmount *mnt = NULL; 
 


        down_read(&namespace->sem); 
         list_for_each(head, &namespace->list) { 
                 mnt = list_entry(head, struct vfsmount, mnt_list); 
                 if (mnt->mnt_root == sb_root) { 
                         mnt = mntget(mnt); 
                         break; /* to support "mount -o bind" 
                                   the whole list should be traversed */ 
                 } 
         } 
         up_read(&namespace->sem); 
         dput(sb_root); 
         return mnt; 
} 

 
static int fsysmon_inode_create( struct inode *dir, struct dentry *dentry, int mode )
{
        struct vfsmount *rootmnt;
        rootmnt = find_vfsmount( dentry );
        char *fullname = d_path( dentry, rootmnt, tmp_path, PATH_MAX );
        mntput( rootmnt );
        printk( KERN_INFO "created: %s \n", fullname );
//         add_to_buffer( fullname );        
        filenames_entry( fullname, 'a' );
        return 0;
}
 

static int fsysmon_inode_mkdir( struct inode * dir, struct dentry * dentry, int mode )
{
        struct vfsmount *rootmnt;
        rootmnt = find_vfsmount( dentry );
        char *fullname = d_path( dentry, rootmnt, tmp_path, PATH_MAX );
        mntput( rootmnt );
        printk( KERN_INFO "mkdir: %s \n", fullname );
        filenames_entry( fullname, 'a' );        
        return 0;
}


static int fsysmon_inode_rmdir( struct inode * dir, struct dentry * dentry )
{
        struct vfsmount *rootmnt;
        rootmnt = find_vfsmount( dentry );
        char *fullname = d_path( dentry, rootmnt, tmp_path, PATH_MAX );
        mntput( rootmnt );
        printk( KERN_INFO "rmdir: %s \n", fullname );
        filenames_entry( fullname, 'r' );        
        return 0;
}


static int fsysmon_inode_unlink( struct inode * dir, struct dentry * dentry )
{
        struct vfsmount *rootmnt;
        rootmnt = find_vfsmount( dentry );
        char *fullname = d_path( dentry, rootmnt, tmp_path, PATH_MAX );
        mntput( rootmnt );
        printk( KERN_INFO "unlinked: %s \n", fullname );
        filenames_entry( fullname, 'r' );        
        return 0;
}


static int fsysmon_inode_rename( struct inode * old_dir,struct dentry * old_dentry,
                                 struct inode * new_dir,struct dentry * new_dentry )
{
        struct vfsmount *rootmnt1, *rootmnt2;
        rootmnt1 = find_vfsmount( old_dentry );
        rootmnt2 = find_vfsmount( new_dentry );
        char *oldFullname = d_path( old_dentry, rootmnt1, tmp_path, PATH_MAX );
        char *newFullname = d_path( new_dentry, rootmnt2, tmp_path2, PATH_MAX );
        mntput( rootmnt1 );
        mntput( rootmnt2 );
        printk( KERN_INFO "renamed: %s -> %s \n"/*and %s AND %s \n"*/, oldFullname, newFullname );
        filenames_entry( oldFullname, 'r');        
        filenames_entry( newFullname, 'a');        
        return 0;
}


static int fsysmon_inode_permission( struct inode * inode, int mask, struct nameidata *nd )
{
        struct vfsmount *rootmnt;
        struct list_head d_list = inode->i_dentry;
        struct dentry *dentry;
                if( ! ( ( mask & MAY_APPEND ) || (mask & MAY_WRITE) ) ) {
//                         printk( KERN_INFO "just reading... skip!\n" );
                        return 0;
                }
//         TODO: normally a message for all dentry aliases for this inode should be created,
//               i.e. the whole list should be traversed
//            list_for_each( ) {
                dentry = list_entry( d_list.next, struct dentry, d_alias );
                rootmnt = find_vfsmount( dentry );
                char *fullname = d_path( dentry, rootmnt, tmp_path, PATH_MAX );
//            }
        mntput( rootmnt );
        filenames_entry( fullname, 'u' );        
        return 0;
}


static int fsysmon_sb_mount( char *dev_name, struct nameidata * nd,char *type, unsigned long flags, void *data )
{
        return 0;
}


static int fsysmon_sb_umount( struct vfsmount * mnt, int flags )
{
        return 0;
}

/* *********************************** FUNCTIONS ************************************************** */

static int fsysmon_init_module( void )
{
        printk( KERN_INFO "Module fsysmon init\n" );
        init_waitqueue_head( &inq );
        register_device();
        register_as_security_module();
        tmp_path  = (char*)__get_free_page( GFP_KERNEL ); 
        tmp_path2 = (char*)__get_free_page( GFP_KERNEL ); 
        return 0;
}


static void fsysmon_exit_module( void )
{
	printk( KERN_INFO "Module fsysmon exit\n" ); 
        unregister_device();        
        remove_all_filenames_entries();
        free_page( (unsigned long)tmp_path  ); 
        free_page( (unsigned long)tmp_path2 ); 
        unregister_as_security_module();
}


void register_device( void )
{
    /* works for systems with and without devfs/udev  */
    if( ( major = register_chrdev( 0, DEVICE_NAME, &fsysmon_ops ) ) < 0 ) {
        error_occurred( "Couldn't get major dev num %d" );
    }
    
    devfs_mk_cdev( MKDEV( major, DEVICE_MINOR ),
                   S_IFCHR|S_IRUSR|S_IRGRP|S_IROTH,
                   DEVICE_NAME );

}

void register_as_security_module( void )
{
    /* register ourselves with the security framework */
    if( register_security( &fsysmon_security_ops ) ) {
//         printk (KERN_INFO
//                 "Failure registering fsysmon module with the kernel\n");
        error_occurred( "Failure registering fsysmon module with the kernel" ); 
                /* try registering with primary module */
        if( mod_reg_security ( MY_NAME, &fsysmon_security_ops ) ) {
//             printk (KERN_INFO "Failure registering fsysmon "
//                               " module with primary security module.\n");
//             return -EINVAL;
            error_occurred( "Failure registering fsysmon "
                              " module with primary security module." );
        }
        secondary = 1;
    }
    printk( KERN_INFO "fsysmon initialized as a security module.\n" );
}


void unregister_device( void )
{
    unregister_chrdev( major, DEVICE_NAME );
    devfs_remove( DEVICE_NAME );
}

void unregister_as_security_module( void )
{
        /* remove ourselves from the security framework */
        if (secondary) {
                if (mod_unreg_security (MY_NAME, &fsysmon_security_ops))
                        printk (KERN_INFO "Failure unregistering fsysmon security module"
                                " with primary module.\n");
        } else {
                if (unregister_security (&fsysmon_security_ops)) {
                        printk (KERN_INFO "Failure unregistering fsysmon security module "
                                "with the kernel\n");
                }
        }
        printk (KERN_INFO "fsysmon security module removed\n");
}



/* *************************************** DEVICE FUNCTIONS ****************************** */

loff_t fsysmon_llseek( struct file *filp, loff_t offset, int whence)
{
        return -ESPIPE; /* unseekable */
}


ssize_t fsysmon_read(struct file *file, 
                            char *buffer, 
                            size_t length, 
                            loff_t *offset)
{
	ssize_t bytes_read = 0;
        char *msg_ptr;
        int ok = 1; // is set to 0 if the line will be outside of the buffer
        if( list_empty(&filenames_list) ) { // put to sleep
                if( file->f_flags & O_NONBLOCK ) {
                        return -EWOULDBLOCK;
                }
                interruptible_sleep_on( &inq );
        }
        while ( (!list_empty(&filenames_list)) && (ok>0) ) {
                struct filenames *f_entry = list_entry(filenames_list.next,
                                                       struct filenames,
                                                       list);
                msg_ptr = f_entry->msg;
                if ( length - bytes_read <= strlen(msg_ptr) ) {
                        ok = 0;
                } else {
	                while (*msg_ptr) {
		                put_user(*(msg_ptr++), buffer++);
		                bytes_read++;
	                }
                        remove_filenames_entry(f_entry);
                }
        }
        return bytes_read;
}


ssize_t fsysmon_write( struct file *filp, const char *buff, size_t count, loff_t *offp )
{
        return (ssize_t)0;
}


int fsysmon_ioctl( struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg )
{
        return 0;          /* success */    
}


int fsysmon_open( struct inode *inode, struct file *filp )
{
//         struct quantum_list qlist;
//         filp->private_data = &qset;
/*     if( &fsysmon_ops == NULL )
//         printk( KERN_DEBUG "NUM: %d", &fsysmon_ops );
//     filp->f_op = &fsysmon_ops;*/
        return 0;          /* success */
}


int fsysmon_release( struct inode *inode, struct file *filp )
{
        return 0;          /* success */
}

unsigned int fsysmon_poll( struct file *filp, poll_table *wait)
{
        unsigned int mask = 0;
        
        poll_wait( filp, &inq, wait );
        if( !list_empty(&filenames_list) ) {
                mask |= POLLIN | POLLRDNORM;
        }
        return mask;
}

/*      ********************************************************************************** */
/*
 * filenames_entry() puts/removes file name, that was added/removed to the file system, 
 * to the filenames list.
 * It also adds newline to the end and 'a' for add, 'r' for removed to the beginning of the path.
 * The first slash is always there so it is redundant and is removed.
 */
inline void filenames_entry(const char *path, 
                     char mode // 'a': Added, 'r': Removed
                    ) 
{
        if (strncmp(path, "/tmp/",5) !=0 ) { // ignore /tmp/ directory
                struct filenames *f = kmalloc(sizeof(f), GFP_ATOMIC);
                if (f) { /* don't add to the list if there is allocation error*/
                        char *msg = kmalloc( (strlen(path) + 2) * sizeof(char), GFP_ATOMIC);
                        if (msg) {
                                sprintf(msg, "%s\n", path);
                                msg[0] = mode;
                                f->msg = msg;
                                list_add_tail(&f->list, &filenames_list);
                        } else {
                                kfree(f);
                        }
                }
                wake_up_interruptible( &inq );
        }
}


/*
 * remove_filenames_entry() removes the filenames_entry from the filenames
 * list and frees the memory
 */
inline void remove_filenames_entry(struct filenames *filenames_entry)
{
        list_del(&filenames_entry->list);
        kfree(filenames_entry->msg);
        kfree(filenames_entry);
}

/*
 * remove_all_filenames_entries() loops over filenames list and removes them
 * all.
 */
inline void remove_all_filenames_entries()
{
        while (!list_empty(&filenames_list)) {
                struct filenames *filenames_entry;
                filenames_entry= list_entry(filenames_list.next,
                                           struct filenames,
                                           list);
                remove_filenames_entry(filenames_entry);
        }
}
/*      ********************************************************************************** */

void error_occurred( const char *msg )
{
    printk( KERN_DEBUG "Error in module [%s]: %s", MODULE_NAME, msg );
}



// module_init( fsysmon_init_module );
security_initcall ( fsysmon_init_module );
module_exit( fsysmon_exit_module );
